#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import argparse
import scipy.stats as stat
from scipy.integrate import simps
from shutil import copyfile
from sklearn import metrics

from analysis.utils import plot_densities
from seair.seair import SeairOde
from seair.utils import compute_q

#################
# CONSTANTS
#################
MIN_0 = 0.0
MAX_100 = 100.0
MIN_05 = 0.5
MAX_1 = 1.0
AUC_COMPUTE_SIZE = 500
ALPHA_N_BETA_P = 2
ALPHA_P_BETA_N_MAX = 1000
DAYS_AFTER_COMPLETE_RELEASE = 200


#################
# CDFs Definitions
#################
def cdf_beta_n(alpha_n, beta_n):
    def inner_cdf_beta_n(t):
        return stat.beta.cdf(t, alpha_n, beta_n)
    return inner_cdf_beta_n


def cdf_beta_p(alpha_p, beta_p):
    def inner_cdf_beta_p(t):
        return stat.beta.cdf(t, alpha_p, beta_p)
    return inner_cdf_beta_p


#################
# Reverse engineer betas parameters from AUC value
#################
def compute_auc(alpha_p, beta_p, alpha_n, beta_n):

    a = np.linspace(0, 1, AUC_COMPUTE_SIZE)

    q_FP = np.zeros((AUC_COMPUTE_SIZE,), dtype=float)
    q_TP = np.zeros((AUC_COMPUTE_SIZE,), dtype=float)

    for i in range(AUC_COMPUTE_SIZE):
        q_FP[i] = 1-stat.beta.cdf(a[i], alpha_n, beta_n)
        q_TP[i] = 1-stat.beta.cdf(a[i], alpha_p, beta_p)

    return metrics.auc(q_FP, q_TP)


def auc_to_betaparameters(auc, fixed_param):

    if auc == 0.5:
        return fixed_param
    if auc == 1.0:
        return ALPHA_P_BETA_N_MAX

    fct_temp = lambda param: compute_auc(param, fixed_param, fixed_param, param)-auc
    g, d = fixed_param, 10.
    eps = 10**(-1)
    while d-g > eps:
        m = (g+d)/2
        if fct_temp(m) == 0:
            return m
        elif fct_temp(g) * fct_temp(m) < 0:
            d=m
        else:
            g=m
    return (g+d)/2


#################
# Argparse validation
#################
def float_range_0_100_type(arg):
    try:
        f = float(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be a floating point number")
    if f < MIN_0 or f > MAX_100:
        raise argparse.ArgumentTypeError("Argument must be <= " + str(MAX_100) + "and >= " + str(MIN_0))
    return f


def float_range_05_1_type(arg):
    try:
        f = float(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be a floating point number")
    if f < MIN_05 or f > MAX_1:
        raise argparse.ArgumentTypeError("Argument must be <= " + str(MAX_1) + "and >= " + str(MIN_05))
    return f


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run SEAIR model given all the parameters')
    # Initial conditions
    parser.add_argument("N_0", help="Total number of individuals at time 0", type=int, default=67*10**6)
    parser.add_argument("R_0", help="Total number of immune at time 0", type=int, default=4*10**6)
    parser.add_argument("Inf_0", help="Total number of infected E + A + I at time 0", type=int, default=20*10**4)
    parser.add_argument("ICU_0", help="Total number of individuals in ICU at time 0", type=int, default=4*10**3)
    parser.add_argument("I_max", help="Healthcare capacity", type=int, default=10000)

    # Hypothesis
    parser.add_argument("p", help="Proportion of people with mild symptoms", type=float_range_0_100_type, default=99.5)
    parser.add_argument("delta_r", help="To be thought of how much released people have lowered their exits from home", type=float_range_0_100_type, default=20.0)
    parser.add_argument("delta_c", help="To be thought of how much people in lockdown have lowered their exits from home", type=float_range_0_100_type, default=60.0)

    # Release strategy
    parser.add_argument('T1', help="Date of second partial release (in days after T0)", type=int, default=300)
    parser.add_argument('T2', help="Date of complete release (in days after T0)", type=int, default=600)
    parser.add_argument("prop0", help="Ratio of released population at time t=0", type=float_range_0_100_type, default=70.0)
    parser.add_argument("prop1", help="Cumulated ratio of released population at the start of the second period (time T1+1)", type=float_range_0_100_type,
                        default=90.0)

    # Population distribution
    parser.add_argument('risk_auc', help="Discriminative capacity of the risk score", type=float_range_05_1_type, default=0.7)

    args = parser.parse_args()

    # Initial conditions
    N_0 = args.N_0
    R_0 = args.R_0
    Inf_0 = args.Inf_0
    ICU_0 = args.ICU_0
    I_max = args.I_max

    # Hypothesis
    p = args.p * 1.0 / 100.0
    delta_r = args.delta_r * 1.0 / 100.0
    delta_c = args.delta_c * 1.0 / 100.0

    # Release strategy
    T = np.array([args.T1, args.T2, args.T2+DAYS_AFTER_COMPLETE_RELEASE])
    prop = np.array([args.prop0 * 1.0 / 100.0, args.prop1 * 1.0 / 100.0, 1.0])
    print(T, prop)

    # Population distribution
    risk_auc = args.risk_auc
    alpha_n = ALPHA_N_BETA_P
    beta_p = ALPHA_N_BETA_P
    alpha_p = auc_to_betaparameters(risk_auc, ALPHA_N_BETA_P)
    beta_n = alpha_p

    # Contact
    grid_prop = np.linspace(0, 1, 1000)
    av_r = 1 - (1 - delta_r) * ((1 - delta_r) * grid_prop + (1 - delta_c) * (1 - grid_prop))
    av_c = 1 - (1 - delta_c) * ((1 - delta_r) * grid_prop + (1 - delta_c) * (1 - grid_prop))
    plt.plot(grid_prop, av_r, 'r', label='Released')
    plt.plot(grid_prop, av_c, 'b', label='Locked Down')
    plt.legend(loc='best')
    plt.title('Contacts : Impact of lower contact rates \ngiven the ratio of the population released')
    plt.xlabel('Cumulated released population ratio')
    plt.ylabel('Contact Reduction ratio')
    plt.grid()
    plt.savefig('./contacts.png')

    # Densities of released and non-released people
    plt.clf()
    plot_densities(alpha_p, beta_p, alpha_n, beta_n, './')

    # SEAIR Model

    ## Given population characteristics, generate proportion of each population to be released
    q = compute_q(p, cdf_beta_p(alpha_p, beta_p), cdf_beta_n(alpha_n, beta_n), prop)

    ## Run SEAIR model
    seair = SeairOde(p=p,
                     N_0=N_0,
                     R_0=R_0,
                     Inf_0=Inf_0,
                     ICU_0=ICU_0)
    [t, sol] = seair.progressive_release(T, q, delta_r, delta_c)

    ## Plot number of deaths
    plt.clf()
    plt.plot(t, sol[:, SeairOde.get_y_order().index("D")], 'b')
    plt.xlabel('t')
    plt.ylabel('Number of deaths')
    plt.title('SEAIR response of number of deaths')
    plt.grid()
    plt.savefig('./model_response_deaths.png')

    ## Plot severe infections
    plt.clf()
    plt.stackplot(t,
                  np.vstack([sol[:, SeairOde.get_y_order().index("I_sr")],
                             sol[:, SeairOde.get_y_order().index("I_sc")]]),
                  labels=['Severe Infections from\nLow-Risk', 'Severe Infections from\nHigh-Risk'])
    plt.xlabel('t')
    plt.ylabel('Number severe infections')
    plt.legend(loc='upper right')
    plt.title('SEAIR response for severe infections\n(Stacked)')
    plt.grid()
    plt.savefig('./model_response_severe_infections.png')

    ## Plot saturation of hospitals
    plt.clf()
    plt.stackplot(t,
                  np.vstack([sol[:, SeairOde.get_y_order().index("ICU_r")],
                             sol[:, SeairOde.get_y_order().index("ICU_c")]]),
                  labels=['ICU from\nLow-Risk', 'ICU from\nHigh-Risk'])
    plt.title('Saturation of ICU\n(stacked)')
    plt.ylabel('Number of people in ICU')
    plt.plot(t, I_max * np.ones(t.size), 'brown', label='ICU Saturation')
    plt.legend(loc='upper right')
    plt.xlabel('t')
    plt.grid()
    plt.savefig('./model_response_hospital_saturation.png')

    ## Plot percentage of removed (immunised)
    plt.clf()
    plt.stackplot(t,
                  np.vstack([sol[:, SeairOde.get_y_order().index("R_mr")] / seair.get_N_0(),
                             sol[:, SeairOde.get_y_order().index("R_mc")] / seair.get_N_0()]),
                  labels=['Low Risk - Mild Infections', 'High Risk - Mild Infections'])
    plt.ylabel('Stacked Ratio of total population')
    plt.legend(loc='lower right')
    plt.title(
        'SEAIR response for recovered people\n(People infected by the disease, survived and are now immune)\nMild Infections',
        fontsize=10)
    plt.xlabel('t')
    plt.grid()
    plt.savefig('./model_response_recovered_mild_symptoms.png')

    plt.clf()
    plt.stackplot(t,
                  np.vstack([sol[:, SeairOde.get_y_order().index("R_sr")] / seair.get_N_0(),
                             sol[:, SeairOde.get_y_order().index("R_sc")] / seair.get_N_0()]),
                  labels=['Low Risk - Severe Infections', 'High Risk - Severe Infections'])
    plt.ylabel('Stacked Ratio of total population')
    plt.legend(loc='lower right')
    plt.title(
        'SEAIR response for recovered people\n(People infected by the disease, survived and are now immune)\nSevere Infections',
        fontsize=10)
    plt.xlabel('t')
    plt.grid()
    plt.savefig('./model_response_recovered_severe_symptoms.png')

    ## Compute some indicators for a posteriori analysis
    ICU = sol[:, SeairOde.get_y_order().index("ICU_c")] + sol[:, SeairOde.get_y_order().index("ICU_r")]
    excess = ICU - I_max
    ## Boolean : do we exceed the ICU capacity I_max
    ICU_exceeded = np.sum(excess > 0.0) > 0.0
    ## Integrate the excess need for ICU beds
    excess[excess < 0] = 0
    excess_area = simps(excess, np.arange(excess.shape[0]))/500 # 500 because of 500 time discretization
    ## Get max of excess
    max_excess = np.max(excess)
    ## Write in file for IPOL to be able to interpret
    with open("algo_info.txt", "w") as f:
        f.write(f"ICU_exceeded={'Yes' if ICU_exceeded == True else 'No'}\n")
        f.write(f"excess_area={int(excess_area)}\n")
        f.write(f"max_excess={int(max_excess)}\n")
    copyfile("algo_info.txt", "res.txt")
